/** @license 2022 Neil Edelman, distributed under the terms of the
 [MIT License](https://opensource.org/licenses/MIT).

 Scan journal entries for score. */

#if 0
#include "../src/journal.h"
#include "../src/scan_score.h"
#include <inttypes.h> /* C99 */
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <limits.h>
#endif

#include "../src/journal.h"
#include "../src/scores.h"
#include <stdio.h>
//#include <string.h>
//#include <stdlib.h>
#include <assert.h>


/* One array. */
static void scorelist_to_string(const struct score *const s,
	char (*const z)[12]) {
	const char *a = s->name.a, *b;
	char *y = *z;
	b = s->name.b <= a + 11 ? s->name.b : a + 11;
	while(a < b) *(y++) = *(a++);
	*y = '\0';
}
#define ARRAY_NAME scorelist
#define ARRAY_TYPE struct score
#define ARRAY_TO_STRING
#define ARRAY_BODY
#include "../src/array.h"


/* Tree mapping from date-line to indices. */
static void score_to_string(const union line64 line, const size_t *const u,
	char (*const a)[12]) { (void)u; date32_to_string(line.date, a); }
static int score_compare(const union line64 a, const union line64 b)
	{ return a.u64 > b.u64; }
#define TREE_NAME score
#define TREE_KEY union line64
#define TREE_VALUE size_t
#define TREE_COMPARE
#define TREE_TO_STRING
#define TREE_DEFAULT 0
#define TREE_BODY
#include "../src/tree.h"


/*!conditions:re2c*/

static int scan(union date32 date, const char *const buffer,
	struct scores *const scores) {
	const char *YYCURSOR, *YYMARKER, *yyt1, *yyt2, *s0, *s1;
	enum YYCONDTYPE condition = yycline;
	size_t line = 1;
	char datestr[12] = {0};
	const char *why = "unexpected";
	struct score *score = 0;
	assert(buffer && scores);
	YYCURSOR = YYMARKER = yyt1 = buffer;
	/*!re2c /**/
	re2c:define:YYCTYPE = char;
	re2c:yyfill:enable = 0;
	re2c:define:YYGETCONDITION = "condition";
	re2c:define:YYSETCONDITION = "condition = @@;";
	re2c:define:YYGETCONDITION:naked = 1;
	re2c:define:YYSETCONDITION:naked = 1;

	ws = [ \t];
	glyph = [^\x00-\x20\x7f]; // [^\x00\n\t ] + all weird
	semitext = glyph \ ";";
	natural = [1-9][0-9]*;
	uint = [0-9]+;
	keyword = [A-Za-z0-9][A-Za-z0-9_-]*;
	date = natural "-" [0-1][0-9] "-" [0-3][0-9];
	*/
	for( ; ; ) { /*!re2c /**/
		/* Default ignore. */
		<skip> [^\n\x00] { continue; }
		<skip> "\x00" { why = "no newline at end of file"; goto catch; }
		<line> "\x00" { return 1; }
		<line, skip> "\n" => line { line++; continue; }
		<line> * :=> skip
		<line> "::" / [^:] :=> score

		<score> * { why = "score unrecognized"; goto catch; }
		/* Already there. Use the map to get the index from the keyword and
		 then stick a marker in the tree with that index. */
		<score> @s0 keyword @s1 / "\n" => skip { also_add_to_tree: {
			const struct pair keyword = pair(s0, s1);
			const union line64 key = { { (uint32_t)line, date } };
			size_t idx, *pidx;
			if(line > UINT32_MAX)
				{ errno = ERANGE; why = "too many lines of text"; goto catch; }
			if(!(idx = pair_map_table_get(&scores->map, keyword)))
				{ why = "keyword not introduced"; goto catch; }
			/* fixme: bulk */
			switch(score_tree_try(&scores->dates, key, &pidx)) {
			case TREE_PRESENT: why = "duplicate key"; /* _Sic_. */
			case TREE_ERROR: goto catch;
			case TREE_ABSENT: *pidx = idx; break;
			}
			date32_to_string(date, &datestr);
			printf("%s: <%.*s>\n", datestr, (int)(s1 - s0), s0);
			continue;
		} }
		/* New score. */
		<score> @s0 keyword @s1 ":" => score_name {
			size_t *idx;
			switch(pair_map_table_assign(&scores->map, pair(s0, s1), &idx)) {
			case TABLE_PRESENT: errno = EDOM; why = "new keyword already used";
			case TABLE_ERROR: goto catch; /* _Sic_. */
			case TABLE_ABSENT: *idx = 0; break;
			}
			if(!(score = scorelist_array_new(&scores->list))) goto catch;
			*idx = (size_t)(score - scores->list.data);
			/*struct pair key, name; union date32 date; unsigned edges;*/
			score->key.a = s0, score->key.b = s1;
			score->name.a = 0, score->name.b = 0;
			score->date.u32 = 0;
			score->edges = 0;
			date32_to_string(date, &datestr);
			fprintf(stderr, "%s: new score <%.*s> stored in list at %zu.\n",
				datestr, (int)(s1 - s0), s0, *idx);
			goto also_add_to_tree;
		}
		/*ws* @name0 [^\x00\n;]+ @name1 ";"
		ws* @year0 natural @year1 "-" @month [0-1][0-9] "-" @day [0-1][0-9] ";"
		ws*
		/ "\n" => skip*/
		<score_name> * { why = "name unrecognized"; goto catch; }
		<score_date> * { why = "date unrecognized"; goto catch; }
		<score_edges> * { why = "edges unrecognized"; goto catch; }
		<score_name> ws* @s0 semitext+ (" " semitext+)* @s1 /* ws* */ ";"
			=> score_date {
			assert(score);
			score->name.a = s0, score->name.b = s1;
		}
		<score_date> ws* "~"? @s0 date ws* ";" => score_edges {
			assert(score);
			if(!pair_to_date(s0, &score->date)) goto catch;
		}
		<score_edges> ws* "~"? @s0 uint @s1 ws* / "\n" => skip {
			assert(score);
			if(!pair_to_natural(s0, s1, &score->edges)) goto catch;
			score = 0; /* Done. */
		}
	*/ }
	assert(0); /* Never gets here. */
catch:
	if(!errno) errno = EILSEQ;
	date32_to_string(date, &datestr);
	fprintf(stderr, "%s line %zu: %s.\n", datestr, line, why);
	return 0;
}

void scores_(struct scores *const s) {
	if(!s) return;
	score_tree_(&s->dates);
	pair_map_table_(&s->map);
	scorelist_array_(&s->list);
}

struct scores scores(struct journal *const j) {
	struct scores s
		= { scorelist_array(), pair_map_table(), score_tree() };
	struct journal_iterator it;
	union date32 k;
	const char *v;
	assert(j);
	{ /* Null is the first item for convenience, (TABLE_DEFAULT). */
		struct score *nul;
		if(!(nul = scorelist_array_new(&s.list))) goto catch;
		nul->key.a = nul->key.b = nul->name.a = nul->name.b = 0;
		nul->date.u32 = 0;
		nul->edges = 0;
	}
	it = journal_iterator(j);
	while(journal_next(&it, &k, &v)) if(!scan(k, v, &s)) goto catch;
	fprintf(stderr, "List of scores: %s.\n"
		"Mapped to indices: %s.\n"
		"Date-line tree: %s.\n", scorelist_array_to_string(&s.list),
		pair_map_table_to_string(&s.map), score_tree_to_string(&s.dates));
	goto finally;
catch:
	scores_(&s);
finally:
	return s;
}

int scores_is_empty(const struct scores *const s)
	{ return !s || !s->dates.root.node; }

int main(void) {
	struct journal j = journal();
	if(journal_is_empty(&j)) goto catch;
	struct scores s = scores(&j);
	if(scores_is_empty(&s)) goto catch;
	goto finally;
catch:
	scores_(&s);
finally:
	return EXIT_SUCCESS;
}
